/*
  IC-22S Arduino replacing Diode Matrix

  This implementation of the code does not require any radio
  modifications aside the 9 Volt rail to the Duplex cercuit.
  Refer to Drawring and images for details.
  In recognition of original concept and work done by VK3ZYZ
  The original code whose encoder routine is implemented here,
  as well as the Oled display format.
  mod VK3ZYZ to add CTCSS 20221004
  Implemented logic to setup Frequency and CTCSS frequency code
  On boot mode is frequency select (Loop 0)
  Oled is oly updated if data changes
  Push switch and mode is now in CTCSS select
  Push again and back to frequency mode
  if no CTCSS is selected display remains blank at CTCSS location
  if a tone is selected the word Tone is shown
  To turn tone off selecttone and rotate encoder to the word No Tone
  tone is now off
  Implemented code to read Duples switch condition using analogue
  voltages to return the three states Duplex A Duplex B and Simples
  This could require radio slected values.
  Fixed Logic to turn off CTCSS when No Tone and in RX mode
*/


#define Version "VK2RK 20221009V2.2"
#include "Adafruit_SSD1306.h"
#define SCREEN_WIDTH 128   // OLED display width, in pixels
#define SCREEN_HEIGHT 32   // OLED display height, in pixels
#define OLED_RESET     -1  // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// encoder definitions
#define ENCODER_A     A0
#define ENCODER_B     A1
#define TRX           A2           // PC2 input HIGH = TX, LOW = RX.
#define ENC_SWITCH    A3           // the push switch on the encoder. 

#define RadDup        A7

#define ENCODER_INPUT_PORT PINC

float SubTone     =    0;
int   SubIndex    =    0;
int   OldSubIndex =    0;
#define RefFrequ      500000    // 16Mhz / 32 steps for CTCSS sine wave generator
int   SineDivider = 4065;       // default sine wave divider
int   SubToneList []            // All the CTCSS frequencies.
{ 0,  670,  693,  719,  744,  770,  797,  825,  854,  885,  915,  948,  974,
  1000, 1035, 1072, 1109, 1148, 1188, 1230, 1273, 1318, 1365, 1413, 1462, 1500,
  1514, 1567, 1598, 1622, 1655, 1679, 1713, 1738, 1773, 1799, 1835, 1862, 1899,
  1928, 1966, 1995, 2035, 2065, 2107, 2181, 2257, 2291, 2336, 2418, 2503, 2541,
};

int sinetable[]                   // one cycle sine wave via R2R network.
{ 0, 1, 2, 3, 5, 7, 10, 13, 16, 19, 22, 25, 27, 28, 29, 30, 31, 30, 29, 28, 27, 25, 22, 19, 16, 13, 10, 7, 5, 3, 2, 1, };

int encoder = 1;
int long Frequency = 146000;                              //Starting frequenccy
int Ch;
float OutFrequency;
int Channel;
int Refresh;                                              //Oled Display Refresh
float DuplexVal;
int Duplex;
float OldDuplexVal;
int period = 1;                                           // Polling Timer
unsigned long micro_now = 1;                              // uS tick
int MicroTick;                                            // Tick Count
int Loop = 0;                                             // Frequency control Mode if 1 in PL control Mode
int PL = 1;
int Break;
int i;
int Tx;
int FR_Disp;
int FR_Disp_Old;
//*******************************************************************************************************************************************
void setup() {

  //***************************************
  //Serial.begin(9600); // can corrupt PLL
  // Serial.flush();
  //***************************************

  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {       // Address may be 0x3D
    for (;;); // Don't proceed, loop forever
  }
  display.clearDisplay();
  display.display();                                      // Update screen
  display.setTextSize(1);                                 // Normal 1:1 pixel scale
  display.setTextColor(WHITE, BLACK);                     // Draw white text
  display.setCursor(0, 0);                                // Start at top-left corner
  display.println(F(Version));
  display.display();                                      // Update screen
  delay(1000);
  display.clearDisplay();
  display.display();                                      // Update screen
  delay(100);
  display.setCursor(0, 0);
  display.print("Ch                    ");
  display.display(); // Update screen
  pinMode(ENCODER_A, INPUT_PULLUP);                       // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);
  pinMode(ENC_SWITCH, INPUT_PULLUP);                      // Encoder push switch
  pinMode(TRX, INPUT);
  pinMode(TRX, INPUT);                                    // Hi = TX, LOW =  RX
  pinMode(RadDup, INPUT);
  DDRB  = B11111;                                         // sets Arduino pins 8 to 12 as outputs.
  PORTB = B10000;                                         // R2R out at rest to half level.
  DDRD  = B11111111;                                      // Radio Matrix Output D0 to D12
  OutFrequency = 146000;                                  // Set Radio to start frequency and Channel number
  Ch = 41;                                                // Ch for starting frequency
  Channel = 41;                                           // Display starting Channel Number
  ProgramRadio();                                         // Program Radio
  DisplayData();                                          // Set Channel Number on display
  PL = 1;                                                 // Initialise Push Switch state
  DuplexMode();                                           // Display Duplex Status
}

//*******************************************************************************************************************************************
void loop() {

  PollSwitches();
  if (PL == 0 && Loop == 0) {                             // Loop 0 Frequerncy loop
    delay(200);
    PL = digitalRead(ENC_SWITCH);
    if (PL == 1 ) {
      // PL = 1;
      Loop = 1;
    }
  }
  if (PL == 0 && Loop == 1) {                             // Loop 1 CTCSS Loop
    delay(200);
    PL = digitalRead(ENC_SWITCH);
    if (PL == 1 ) {
      //PL = 1;
      Loop = 0;
    }
  }
  if (Loop == 0) {                                        // Read Duplex Mode Status
    Matrix_Val();
    if (DuplexVal != OldDuplexVal) {                      // Determine Simplex or Duplex
      DuplexMode();
      Refresh = 1;
      DisplayData();
    }
  }
  Tx = digitalRead(TRX);
  if (Tx == 1) {
    CTCSS_Tone();
  }
  if (Loop == 1) {                                        // Set up CTCSS Frequency
    Refresh = 1;
    DisplayData();
    SetCTCSS();
    delay(250);
  }
}
//*******************************************************************************************************************************************
int read_encoder() {

#define DELAY 100                                         // number of polling-loops   
  static unsigned short integrator;                       // encoder debounce
  static char encoder[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0}; // SWAPPED PHASES
  static unsigned char encoder_index = 0;
  static long count = 0;                                  // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                             // return value
  encoder_index <<= 2;                                    // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);           // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];
  // Debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {                 // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {
      integrator++;
    }
  } else {                                                // encoder not resting on indent
    if (integrator > 0) {
      integrator--;
    }
  }
  // prepare the "return" value
  if (integrator >= DELAY) {                              // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                          // both readings now hold current "count"
    integrator = 0;                                       // reset integrator
  } else {
    value = 0;
  }
  return (value);                                         // Return value to calling function
}

//*******************************************************************************************************************************************
void Matrix_Val() {

  int KHz;

  encoder = read_encoder();
  if (encoder != 0) {
    Refresh = 1;
    KHz = (encoder * 25);
    OutFrequency = Frequency + KHz;
    Channel = Ch + encoder;  // One step = 25 KHz
    Channel = constrain(Channel, 1, 119);
    OutFrequency = constrain(OutFrequency, 145000, 147975);
    Frequency = OutFrequency;
    Ch = Channel;
    DisplayData();
    ProgramRadio();
  }
}

//*******************************************************************************************************************************************
void ProgramRadio() {

  Refresh = 1;
  int PLLdivider = ((OutFrequency - 144400) / 25);
  PORTD = (PLLdivider);
}

//*******************************************************************************************************************************************
void DisplayData() {

  float DispFrequency;
  if (Loop == 0 && Refresh == 1) {
    display.setTextSize(1);                               // Normal 1:1 pixel scale
    display.setCursor(18, 0);
    display.print("   ");                                 // clear the old number
    display.setCursor(18, 0);
    display.print(Channel);                               // display the channel number
    DispFrequency = OutFrequency / 1000;                  // Fix data to display decimal
    if (Tx = 1 && FR_Disp == 1) {                             // Duplex A Frequency fix
      DispFrequency = DispFrequency + 0.600;
    }
    if (Tx = 0 && FR_Disp == 1) {
      DispFrequency = DispFrequency;
    }

    display.setCursor(0, 10);
    display.setTextSize(3);                               // Large  pixel scale
    display.print(DispFrequency, 3);                      // write Frequency
    display.setTextSize(1);                               // Normal 1:1 pixel scale
    display.setCursor(35, 0);                             // write mode to the LCD
    if (Duplex == 1 && Loop == 0) {                                   // Duplex Mode
      display.print("-600 ");
    }
    if (Duplex == 2 && Loop == 0) {                                   // Duplex Mode
      display.print("+600 ");
    }
    if (Duplex == 3 && Loop == 0) {                       // Simplex Mode
      display.print(" SIM ");
    }
    display.setCursor(68, 0);
    display.print("          ");
    if (SubTone > 0 && Loop == 0) {
      display.setCursor(85, 0);
      display.print("Tone");
    }
    display.display();                                    // Update Display
    Refresh = 0;                                          // Turn Off refresh
  }
  if (Loop == 1 && Refresh == 1) {
    display.setCursor(68, 0);                             // set the cursor loacation
    display.print("         ");                           // preclear the old stuff
    display.setCursor(68, 0);                             // yes, set the cursor loacation
    display.print("Tone ");                               // write "tone"
    display.display();
    if (SubTone != 0) {
      if ((SubTone) < 100) {                              // pad leading spacee if needed
        display.print(" ");                               // space if needed
      }
      display.print(SubTone, 1);                          // show the Tone frequency to 1 DP
    }
    if (SubTone == 0) {
      display.setCursor(68, 0);
      display.print("         ");                         // preclear the old stuff
      display.setCursor(68, 0);
      display.print("          ");
      display.setCursor(70, 0);
      display.print("No Tone");
    }
    display.display();                                    // Update OLED
    Refresh = 0;
  }
}
//*******************************************************************************************************************************************
void DuplexMode() {    // Test if radio duplex switch is active

  Duplex_Mode_Sense();
  if (FR_Disp == 1) {
    Duplex = 1;  // -600
  }
  if (FR_Disp == 2) {
    Duplex = 2;  // +600
  }
  if (FR_Disp == 3) {
    Duplex = 3;  // Sim
  }
  Refresh = 1;
  DisplayData();
  OldDuplexVal = DuplexVal;
}

//*******************************************************************************************************************************************
void PollSwitches() {                    // Poll status of duplex switch

  if ((unsigned long)(micros() - micro_now) > period) {
    MicroTick ++;
    micro_now = micros();
  }
  if (MicroTick == 5000) {             // Poll switch every 10mS.
    MicroTick = 0;  // Reset Timer
    DuplexVal = (analogRead(RadDup));   // Read Status of Duplex Switch
    PL = digitalRead(ENC_SWITCH);       // Read Status of Push Switch
    if (PL == 0) {
      Break = 1;
    }
  }
}

//*******************************************************************************************************************************************
void SetCTCSS() {                                         //  select the CTCSS tone

  if (Loop != 0) {
    while (Loop = 1) {                                    // Rcursive call
      float ToneFR;
      encoder = read_encoder();                           // read the encoder
      if (encoder != 0)  {                                // has it turned?
        SubIndex = (SubIndex + encoder);                  // select the CTCSS tone
        SubIndex = constrain(SubIndex, 0, 51);            // Array position "0" Never reached Fix add another "0" to table
        ToneFR  = (SubToneList[SubIndex]);                // table has value * 10 to use int for space saving.
        SubTone  = ToneFR / 10;
        if (SubIndex == 0) {                              // Test if at Array position "0"
          SubTone = 0;
        }
        if (SubIndex != OldSubIndex) {                    // has it changed?
          Refresh = 1;
          DisplayData();                                  // Call screen refresh

          OldSubIndex = SubIndex;                         // record for next time.
          if (SubTone != 0) {                             // do we need a sub tone?
            SineDivider = (RefFrequ / (float)SubTone);    // Yes, int Subtone from table, now = Float/10
          }
          else {
            TIMSK1 = 0;                               // No, subtone off
          }
        }
      }
      PL = digitalRead(ENC_SWITCH);
      if (PL == 0) {                                       // Switch pressed
        PL = 1;
        Loop = 0;
        Refresh = 1;
        DisplayData();
        break;                                             // Exit While Loop
      }
    }
  }
}


//*******************************************************************************************************************************************
void CTCSS_Tone() {

  cli();                                                  // interupsts off for setup
  TCCR1A = 0;
  TCCR1B = 0; //B00000010;                                // no prescaler
  // set compare match register for 1hz increments
  OCR1A = SineDivider ;// = (16*10^6) / (1*1024) - 1 (must be <65536)
  //Serial.print(SineDivider);
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS10 no prescaler
  TCCR1B |= (0 << CS12) | (0 << CS11) | (1 << CS10);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();                                                  // interupsts on
}
ISR (TIMER1_COMPA_vect) {

  if (SubTone != 0 && Tx == 0 ) {
    PORTB = (sinetable[i++]);           // write value from table to PORTB
    if (i >= 32) {
      i = 0;
    }
  }
}

//*******************************************************************************************************************************************
void Duplex_Mode_Sense() {                                    // Convert Analogue Values to Duplex Status

  if (Tx == 1 && DuplexVal == 1023) {
    FR_Disp = 1;                                              // RX Mode A
  }
  if (Tx == 1 && DuplexVal == 0) {
    FR_Disp = 3;                                              // RX Simplex
  }
  if (Tx == 1 && DuplexVal > 165 && DuplexVal < 185) {
    FR_Disp = 2;                                              // RX Mode B
  }
  if (Tx == 0 && DuplexVal > 10 && DuplexVal < 19) {
    FR_Disp = 1;                                              // TX Mode A
  }
  if (Tx == 0 && DuplexVal == 0) {
    FR_Disp = 3;                                              // TX Simplex
  }
  if (Tx == 0 && DuplexVal == 1023) {
    FR_Disp = 2;                                              // TX Mode B
  }

}
